﻿Imports Microsoft.VisualBasic
Imports System
Imports System.Collections.Generic
Imports System.IO
Imports System.Text
Imports System.Windows
Imports System.Windows.Controls
Imports System.Windows.Media

Public Class PageProvider
    Implements IPageProvider
    ' in units of space width
    Private Const FirstLineIndent As Integer = 3
    ' in units of character height
    Private Const ParagraphSpacing As Double = 1.0 / 3
    Private Shared ReadOnly BlackBrush As Brush = New SolidColorBrush(Colors.Black)

    ' Regenerated every time book is loaded
    Private paragraphs As New List(Of ParagraphInfo)()

    Private fontMetrics As New FontMetrics(New Font())
    Private pageSize As Size
    Private fontSize As Double

    ' Only public property
    Public Property Book() As BookInfo

    Protected Sub PreprocessBook(ByVal stream As Stream)
        ' Occurs every time program is run
        GenerateParagraphs(stream)

        ' Occurs only first time program is run
        If Me.Book.Chapters.Count = 0 Then
            GenerateChapters()
        End If
    End Sub

    Private Sub GenerateParagraphs(ByVal stream As Stream)
        Using reader As New StreamReader(stream)
            Dim line As String = Nothing
            Dim prevLine As String = Nothing
            Dim prevIndent As Integer = 0
            Dim paragraphBuilder As New StringBuilder()

            line = reader.ReadLine()
            Do While Nothing IsNot line
                ' Trim line of all space and find the line indent
                line = line.TrimEnd(" "c)
                Dim indent As Integer = line.Length
                line = line.TrimStart(" "c)
                indent -= line.Length

                ' A new paragraph is always implied with a line of zero length
                Dim newParagraph As Boolean = line.Length = 0

                If (Not newParagraph) AndAlso paragraphBuilder.Length > 0 AndAlso prevLine IsNot Nothing Then
                    ' This is the case for a list ("contents" for example)
                    '  where each line has a different indent
                    If indent <> prevIndent Then
                        newParagraph = True

                        ' But some indented text needs to be concatenated (epistoles, for example)
                    ElseIf indent <> 0 Then
                        Dim firstWord As String = line
                        Dim spaceIndex As Integer = line.IndexOf(" "c)

                        If spaceIndex <> -1 Then
                            firstWord = firstWord.Substring(0, spaceIndex)
                        End If

                        ' This works for some books, but not for "Phineas Finn" where
                        '  at least the first letter has shorter lines than normal
                        ' if (indent + prevLine.Length + 1 + firstWord.Length <= 70)
                        '     newParagraph = true
                    End If
                End If

                ' Check for a possible paragraph termination
                If newParagraph Then
                    If paragraphBuilder.Length > 0 Then
                        SubstituteCharacters(paragraphBuilder)

                        ' Create a new ParagraphInfo for the paragraph
                        Dim paragraph As ParagraphInfo = New ParagraphInfo With {.Indent = prevIndent, .Text = paragraphBuilder.ToString()}
                        paragraphs.Add(paragraph)

                        ' Clear the StringBuilder
                        paragraphBuilder.Remove(0, paragraphBuilder.Length)
                    End If

                    If line.Length > 0 Then
                        ' Append the line to the paragraph
                        paragraphBuilder.Append(line)
                    Else
                        ' Add the current line that terminates the paragraph
                        Dim emptyParagraph As ParagraphInfo = New ParagraphInfo With {.Indent = 0, .Text = line}
                        paragraphs.Add(emptyParagraph)
                    End If
                Else
                    ' Insert a space between appended lines assembled into paragraph
                    If paragraphBuilder.Length > 0 Then
                        paragraphBuilder.Append(" ")
                    End If

                    paragraphBuilder.Append(line)
                End If
                prevLine = line
                prevIndent = indent
                line = reader.ReadLine()
            Loop

            ' Add the last paragraph if there's one in progress
            If paragraphBuilder.Length > 0 Then
                SubstituteCharacters(paragraphBuilder)

                Dim paragraph As ParagraphInfo = New ParagraphInfo With {.Indent = prevIndent, .Text = paragraphBuilder.ToString()}
                paragraphs.Add(paragraph)
            End If
        End Using
    End Sub

    ' Does fancy quotes and em-dashes
    Private Sub SubstituteCharacters(ByVal paragraphBuilder As StringBuilder)
        ' Replace boring double quotes with fancy quotes
        Dim isCloseQuote As Boolean = False

        For i = 0 To paragraphBuilder.Length - 1
            If paragraphBuilder(i) = """"c Then
                paragraphBuilder(i) = If(isCloseQuote, ChrW(&H201D), ChrW(&H201C))
                isCloseQuote = isCloseQuote Xor True
            End If
        Next i

        ' Replace double dashes with em dashes separated by spaces
        '  (but only when they're surrounded by non-dashes)
        Dim index As Integer = 0

        Do While index < paragraphBuilder.Length - 1
            ' there isn't a later dash -  there isn't an earlier dash -  next one is dash -  current is dash
            If paragraphBuilder(index) = "-"c AndAlso paragraphBuilder(index + 1) = "-"c AndAlso (index = 0 OrElse paragraphBuilder(index - 1) <> "-"c) AndAlso (index = paragraphBuilder.Length - 2 OrElse paragraphBuilder(index + 2) <> "-"c) Then
                ' double-dash appears alone
                If index = 0 AndAlso paragraphBuilder.Length = 2 Then
                    paragraphBuilder.Remove(index, 2)
                    paragraphBuilder.Insert(index, ChrW(&H2014).ToString())
                    ' double-dash at beginnning of paragraph
                ElseIf index = 0 Then
                    paragraphBuilder.Remove(index, 2)
                    paragraphBuilder.Insert(index, ChrW(&H2014).ToString() & " ")
                    index += 1
                    ' double-dash not followed by at least two characters
                ElseIf index >= paragraphBuilder.Length - 3 Then
                    paragraphBuilder.Remove(index, 2)
                    paragraphBuilder.Insert(index, " " & ChrW(&H2014).ToString())
                    index += 1
                Else
                    paragraphBuilder.Remove(index, 2)
                    paragraphBuilder.Insert(index, " " & ChrW(&H2014).ToString() & " ")
                    index += 2
                End If
            End If

            index += 1
        Loop
    End Sub

    Private Sub GenerateChapters()
        ' Set so the beginning of the book becomes the first chapter
        Dim consecutiveBlankLineCount As Integer = 10
        Dim accumulatedCharacterCount As Integer = 0
        Dim chapterInfo As ChapterInfo = Nothing

        For paragraphIndex = 0 To paragraphs.Count - 1
            If paragraphs(paragraphIndex).Text.Length = 0 Then
                consecutiveBlankLineCount += 1
            Else
                If consecutiveBlankLineCount > 2 Then
                    Dim pageInfo As PageInfo = New PageInfo With {.ParagraphIndex = paragraphIndex, .CharacterIndex = 0, .AccumulatedCharacterCount = accumulatedCharacterCount}

                    chapterInfo = New ChapterInfo With {.Title = If(paragraphs(paragraphIndex).Text.Length < 100, paragraphs(paragraphIndex).Text, paragraphs(paragraphIndex).Text.Substring(0, 100)), .FirstParagraph = paragraphIndex, .ParagraphCount = 1}
                    chapterInfo.Pages.Add(pageInfo)
                    Me.Book.Chapters.Add(chapterInfo)
                Else
                    ' Accumulate number of paragraphs in the chapter
                    chapterInfo.ParagraphCount += 1 + consecutiveBlankLineCount
                End If
                consecutiveBlankLineCount = 0
                accumulatedCharacterCount += paragraphs(paragraphIndex).Text.Length
            End If
        Next paragraphIndex

        ' Flag the last chapter
        Me.Book.Chapters(Me.Book.Chapters.Count - 1).IsLastChapter = True
        Me.Book.CharacterCount = accumulatedCharacterCount
    End Sub

    Public Sub SetPageSize(ByVal size As Size) Implements IPageProvider.SetPageSize
        If pageSize <> size Then
            pageSize = size

            ' TODO: If the page size can really change,
            '      pagination information needs to be invalidated
        End If
    End Sub

    Public Sub SetFont(ByVal fontFamily As FontFamily, ByVal fontSize As Double) Implements IPageProvider.SetFont
        fontMetrics = New FontMetrics(New Font(fontFamily.Source))

        ' Me.fontFamily = fontFamily
        Me.fontSize = fontSize

        ' TODO: If the font and font size can really change,
        '      pagination information needs to be invalidated
    End Sub


    Public Function QueryLastPage(ByVal chapterIndex As Integer, <System.Runtime.InteropServices.Out()> ByRef pageIndex As Integer) As FrameworkElement Implements IPageProvider.QueryLastPage
        Return GetLastPage(chapterIndex, pageIndex, False)
    End Function

    Public Function GetLastPage(ByVal chapterIndex As Integer, <System.Runtime.InteropServices.Out()> ByRef pageIndex As Integer) As FrameworkElement Implements IPageProvider.GetLastPage
        Dim element As FrameworkElement = GetLastPage(chapterIndex, pageIndex, True)
        Return element
    End Function

    Private Function GetLastPage(ByVal chapterIndex As Integer, <System.Runtime.InteropServices.Out()> ByRef pageIndex As Integer, ByVal paginateIfNecessary As Boolean) As FrameworkElement
        If pageSize.IsEmpty Then
            pageIndex = -1
            Return Nothing
        End If

        If chapterIndex < 0 OrElse chapterIndex > Me.Book.Chapters.Count - 1 Then
            pageIndex = -1
            Return Nothing
        End If

        Dim chapter As ChapterInfo = Me.Book.Chapters(chapterIndex)
        Dim lastIndex As Integer = chapter.Pages.Count - 1

        ' IsPaginated should be implied
        If chapter.Pages(lastIndex).IsLastPageInChapter AndAlso chapter.Pages(lastIndex).IsPaginated Then
            pageIndex = lastIndex
            Return BuildPageElement(chapter, chapter.Pages(lastIndex))
        End If

        If Not paginateIfNecessary Then
            pageIndex = -1
            Return Nothing
        End If

        Do While Not chapter.Pages(chapter.Pages.Count - 1).IsLastPageInChapter
            Paginate(chapterIndex, chapter.Pages.Count - 1)
        Loop

        pageIndex = chapter.Pages.Count - 1
        Return BuildPageElement(chapter, chapter.Pages(chapter.Pages.Count - 1))
    End Function

    Public Function GetPage(ByVal chapterIndex As Integer, ByVal pageIndex As Integer) As FrameworkElement Implements IPageProvider.GetPage
        If pageSize.IsEmpty Then
            Return Nothing
        End If

        If pageIndex < 0 Then
            Return Nothing
        End If

        If chapterIndex < 0 OrElse chapterIndex > Me.Book.Chapters.Count - 1 Then
            Return Nothing
        End If

        Dim chapter As ChapterInfo = Me.Book.Chapters(chapterIndex)

        ' Check if the last page of the paragraph has already been established, 
        '  and the requested page is greater than that last page
        If chapter.Pages(chapter.Pages.Count - 1).IsLastPageInChapter AndAlso pageIndex > chapter.Pages.Count - 1 Then
            Return Nothing
        End If

        ' If the last page of a chapter is requested, and it's already available, get it
        If pageIndex < chapter.Pages.Count AndAlso chapter.Pages(pageIndex).IsLastPageInChapter Then
            Return BuildPageElement(chapter, chapter.Pages(pageIndex))
        End If

        ' Otherwise, if the page has already been paginated, get it
        If pageIndex + 1 < chapter.Pages.Count AndAlso chapter.Pages(pageIndex).IsPaginated Then
            Return BuildPageElement(chapter, chapter.Pages(pageIndex))
        End If

        If pageIndex <> chapter.Pages.Count - 1 Then
            Throw New Exception(String.Format("GetPage can only paginate one page per call: pageIndex = {0} while chapter.Pages.Count = {1}", pageIndex, chapter.Pages.Count))
        End If

        Dim pageInfo As PageInfo = chapter.Pages(pageIndex)

        Dim nextPageInfo As PageInfo = Paginate(chapter, pageInfo)

        ' ie, if !pageInfo.IsLastPageInChapter
        If nextPageInfo IsNot Nothing Then
            chapter.Pages.Add(nextPageInfo)
        End If

        Return BuildPageElement(chapter, pageInfo)
    End Function

    Private Function Paginate(ByVal chapterIndex As Integer, ByVal pageIndex As Integer) As Boolean
        If chapterIndex < 0 OrElse chapterIndex > Me.Book.Chapters.Count - 1 Then
            Return False
        End If

        Dim chapter As ChapterInfo = Me.Book.Chapters(chapterIndex)

        If pageIndex < chapter.Pages.Count AndAlso chapter.Pages(pageIndex).IsPaginated Then
            Return True
        End If

        If chapter.Pages(chapter.Pages.Count - 1).IsLastPageInChapter Then
            Return False
        End If

        Dim pageInfo As PageInfo = chapter.Pages(pageIndex)
        Dim nextPageInfo As PageInfo = Paginate(chapter, pageInfo)

        If nextPageInfo IsNot Nothing Then
            chapter.Pages.Add(nextPageInfo)
        End If

        Return True
    End Function

    ' TODO: Could pass chapter and pageIndex here, to obtain pageInfo within, and then can set nextPageInfo
    Private Function Paginate(ByVal chapter As ChapterInfo, ByVal pageInfo As PageInfo) As PageInfo
        ' Get dimensions of space character
        Dim spaceSize As Size = fontMetrics.MeasureText(" ")
        Dim lineHeight As Integer = CInt(Math.Ceiling(fontSize * spaceSize.Height))
        Dim spaceWidth As Integer = CInt(Math.Ceiling(fontSize * spaceSize.Width))

        ' Initialize a couple variables
        Dim accumHeight As Integer = 0
        Dim characterCount As Integer = pageInfo.AccumulatedCharacterCount
        Dim nextPageInfo As PageInfo = Nothing
        Dim paragraphIndex As Integer = pageInfo.ParagraphIndex
        Dim characterIndex As Integer = pageInfo.CharacterIndex

        ' Loop for lines on page
        Do
            Dim paragraph As ParagraphInfo = paragraphs(paragraphIndex)

            ' For blank paragraphs, leave a gap...
            If paragraph.Text.Length = 0 Then
                ' .. but not when it's at the top of the page
                If accumHeight > 0 Then
                    accumHeight += CInt(Math.Ceiling(ParagraphSpacing * lineHeight))
                End If
                ' Normal case for non-blank paragraphs
            Else
                ' This is the amount to indent the entire paragraph
                Dim indent As Integer = spaceWidth * paragraph.Indent

                ' Apply first line indent if paragraph indent is zero, and it's
                '  not the first paragraph in the chapter
                If indent = 0 AndAlso characterIndex = 0 AndAlso paragraph IsNot paragraphs(chapter.FirstParagraph) Then
                    indent = spaceWidth * FirstLineIndent
                End If

                ' Start off accumulated line width with the indent
                Dim accumWidth As Integer = indent

                ' Loop for words within line -- always the same paragraph
                Do
                    ' Find next space in paragraph
                    Dim spaceIndex As Integer = paragraph.Text.IndexOf(" "c, characterIndex)
                    Dim wordCharacters As Integer = (If(spaceIndex = -1, paragraph.Text.Length, spaceIndex)) - characterIndex
                    Dim wordWidth As Integer = CInt(Math.Ceiling(fontSize * fontMetrics.MeasureText(paragraph.Text, characterIndex, wordCharacters).Width))

                    ' Check if the accumulated width exceeds the page width, 
                    '      but make sure to have at least one word.
                    If accumWidth > indent AndAlso accumWidth + wordWidth >= pageSize.Width Then
                        Exit Do
                    End If

                    Dim word As WordInfo = New WordInfo With {.LocationLeft = accumWidth, .LocationTop = accumHeight, .ParagraphIndex = paragraphIndex, .CharacterIndex = characterIndex, .CharacterCount = wordCharacters}
                    pageInfo.Words.Add(word)

                    accumWidth += wordWidth
                    characterCount += wordCharacters + (If(spaceIndex = -1, 0, 1))
                    characterIndex += wordCharacters + (If(spaceIndex = -1, 0, 1))

                    ' End of line because it's the end of the paragraph
                    If spaceIndex = -1 Then
                        characterIndex = 0
                        Exit Do
                    End If

                    ' End of line because no more words will fit
                    If accumWidth + spaceWidth >= pageSize.Width Then
                        Exit Do
                    End If

                    accumWidth += spaceWidth

                    ' Otherwise, point to next word
                    characterIndex = spaceIndex + 1

                    ' ready for next word
                Loop
                ' Bump up the accumulated height of the page
                accumHeight += lineHeight
            End If

            ' Start a new line, but check first if it's a new paragraph
            If characterIndex = 0 Then
                ' If we've been working with the last paragraph in the chapter,
                '      we're also done with the page
                If paragraphIndex + 1 = chapter.FirstParagraph + chapter.ParagraphCount Then
                    pageInfo.IsLastPageInChapter = True
                    Exit Do
                End If
                ' Otherwise, kick up the paragraphIndex
                paragraphIndex += 1
            End If

            ' Check if the next paragraph is too much for this page
            If (paragraphs(paragraphIndex).Text.Length = 0 AndAlso accumHeight + lineHeight \ 2 > pageSize.Height) OrElse accumHeight + lineHeight > pageSize.Height Then
                Exit Do
            End If
        Loop

        pageInfo.IsPaginated = True

        ' Create PageInfo for next page
        If Not pageInfo.IsLastPageInChapter Then
            nextPageInfo = New PageInfo With {.ParagraphIndex = paragraphIndex, .CharacterIndex = characterIndex, .AccumulatedCharacterCount = characterCount}
        End If

        Return nextPageInfo
    End Function

    Private Function BuildPageElement(ByVal chapter As ChapterInfo, ByVal pageInfo As PageInfo) As FrameworkElement
        If pageInfo.Words.Count = 0 Then
            Paginate(chapter, pageInfo)
        End If

        Dim canvas As New Canvas()

        For Each word In pageInfo.Words
            Dim txtblk As TextBlock = New TextBlock With {.FontFamily = fontMetrics.Font.FontFamily, .FontSize = Me.fontSize, .Text = paragraphs(word.ParagraphIndex).Text.Substring(word.CharacterIndex, word.CharacterCount), .Tag = word}

            canvas.SetLeft(txtblk, word.LocationLeft)
            canvas.SetTop(txtblk, word.LocationTop)
            canvas.Children.Add(txtblk)
        Next word
        Return canvas
    End Function
End Class